
open Arith
open Gamma.Lanczos

(* パイプライン演算子 *)
let (|>) x f = f x

(* 分数微分方程式 *)

(* α = 1/2, β = 1/4 : 平均化法 *)
type t = { a : float; b : float; alpha : float; beta : float; q : float }

let trapezoidal_intg q xls xn h =
  let n = List.length xls in
  match xls with 
  | [] -> 0.
  | [ x0 ] -> 0.5 *. h *. (x0 +. xn)
  | _ ->
      begin
	let x0 = List.hd xls in
	let xm =
	  List.tl xls
  |> List.mapi (fun k x -> x *. (((float (n-k)) *. h) ** (-.q)))
  |> List.fold_left (+.) 0.
	in
	0.5 *. h *. (x0 *. (((float n) *. h) ** (-.q)) +. xn *. (h ** (-.q)) +. 2. *. xm)
      end

let simple_newmark_beta fde x0 vls x_pre v_pre a_pre f h =
  let n_smp = List.length vls in
  let t = (float (n_smp + 1)) *. h in
  
  let (a, b) = (fde.a, fde.b) in
  let (al, bt) = (fde.alpha, fde.beta) in
  let r = al /. bt in
  let q = fde.q in
  let gq = gamma (1. -. q) in
  
  let i_0 = x0 /. (t**q) in
  let sum_v =  trapezoidal_intg q vls v_pre h in
  let i_np = 0.5 *. h *. sum_v in 
  
  let l0 = 1. /. (bt *. (h *^2)) in
  let l1 = a *. (h ** (-.q)) *. r /. ((1. -. q) *. (2. -. q) *. gq) in
  let cl = l0 +. l1 +. b in
  
  let rf = (f t) in
  let r1 = -. a /. gq *. (i_0 +. i_np) in
  let r2 = x_pre /. (bt *. (h*^2)) +. v_pre /. (bt *. h) +. a_pre *. (0.5 /. bt -. 1.) in
  let r3 =
    a *. (h ** (1. -. q)) /. ((1. -. q) *. (2. -. q) *. gq) *. (x_pre *. r /. h +. v_pre *. (q -. 2. +. r) +. (0.5 *. r -. 1.) *. h *. a_pre) in
  let cr = rf +. r1 +. r2 +. r3 in
  cr /. cl

let acceleration fde x_cur x_pre v_pre a_pre h  =
  let bt = fde.beta in
  x_cur /. (bt *. (h *^ 2))
    -. x_pre /. (bt *. (h *^ 2))
    -. v_pre /. (bt *. h)
    -. a_pre *. (0.5 /. bt -. 1.)
    
let velocity fde a_cur v_pre a_pre h =
  let al = fde.alpha in
  al *. h *. a_cur +. v_pre +. (1. -. al) *. h *. a_pre
    

(* FDE Solver *)
let simple_solver fde (x0,v0,a0) f h nsmp =
  let rec simple_solver_core (xls,vls,als) idx =
    let n = List.length xls in
    let x_pre = List.nth xls (n-1) in
    let v_pre = List.nth vls (n-1) in
    let a_pre = List.nth als (n-1) in
    let vls_pre = List.rev vls |> List.tl |> List.rev in
    let x_cur = simple_newmark_beta fde x0 vls_pre x_pre v_pre a_pre f h in
    let a_cur = acceleration fde x_cur x_pre v_pre a_pre h in
    let v_cur = velocity fde a_cur v_pre a_pre h in
    
    let xls_cur = List.append xls [x_cur] in
    let vls_cur = List.append vls [v_cur] in
    let als_cur = List.append als [a_cur] in
    if (idx + 1) >= nsmp then (xls_cur, vls_cur, als_cur)
    else simple_solver_core (xls_cur,vls_cur,als_cur) (idx+1)
  in
  let x1 = simple_newmark_beta fde x0 [] x0 v0 a0 f h in
  let a1 = acceleration fde x1 x0 v0 a0 h  in
  let v1 = velocity fde a1 v0 a0 h in
  
  let xls_0 = x0 :: x1 :: [] in
  let vls_0 = v0 :: v1 :: [] in
  let als_0 = a0 :: a1 :: [] in
  simple_solver_core (xls_0,vls_0,als_0) 1
    


let () =
  let pi = 4. *. atan 1. in
  (* D**2 x + a D**(0.5) x + b x = f *)
  let fde = { a = 0.; b = 1.; alpha = 0.5; beta = 0.25; q = 0.5 } in
  let force t = 0. in
  let h = 2. *. pi *. 0.001 in
  let nsmp = 1000 in
  let x0 = 0. in
  let v0 = 1. in
  let a0 = 0. in
  let (xls, vls, als) = simple_solver fde (x0,v0,a0) force h nsmp in
  List.iter (fun x -> print_endline (string_of_float x)) xls

