(*
  Newmark Beta 法 :
 *)
let beta = 0.25
let gamma = 0.5

let n_deg = 3
let n_smp_1sec = 100
let n_sec = 3
let n_smp = n_smp_1sec * n_sec
let dt = 1. /. (float_of_int n_smp_1sec)
let m = 1.
let c = 5.
let k = 1000.

(* 横ベクトルの計算 *)
let ( +$ ) u v =
  Array.init (Array.length u) (fun n -> u.(n) +. v.(n))
let ( -$ ) u v =
  Array.init (Array.length u) (fun n -> u.(n) -. v.(n))
let ( *$ ) c v =
  Array.init (Array.length v) (fun n -> c *. v.(n))

let print_vector v =
  Array.iter (fun x -> print_string ((string_of_float x) ^ ",")) v;
  print_endline ""

(* 行列の計算 *)
let ( +! ) m1 m2 =
  Array.init (Array.length m1) (fun n -> m1.(n) +$ m2.(n))
let ( -! ) m1 m2 =
  Array.init (Array.length m1) (fun n -> m1.(n) -$ m2.(n))
let ( *! ) c m =
  Array.init (Array.length m) (fun n -> c *$ m.(n))

let print_matrix arr2 =
  Array.iter (fun r -> print_vector r) arr2;
  print_endline ""

(* 前進消去法 *)
let fw_elim mat =
  let l = Array.length mat in

  (* j番目の行ベクトルの掃きだし *)
  let fw_elim_row mat_p i j =
    if i >= j then mat_p.(j)
    else
      begin
	let c_ij = mat_p.(j).(i) /. mat_p.(i).(i) in
	mat_p.(j) -$ (c_ij *$ mat_p.(i))
      end
  in

  let rec fw_elim_each mat_pre n =
      if n >= (l-1) then mat_pre
      else
	begin
	  let mat_cur = Array.init l (fw_elim_row mat_pre n) in
	  fw_elim_each mat_cur (n+1)
	end
  in
  fw_elim_each mat 0

(* 後退代入法 *)
let bw_subst mat =
  let l = Array.length mat in
  let bw_subst_base mat_p i j =
    if i = l-1 then mat_p.(i)
    else mat_p.(i) -$ ((mat_p.(i).(j) /. mat_p.(j).(j)) *$ mat_p.(j))
  in

  let bw_subst_row mat_p row i j =
    if j < i then mat_p.(j) -$ ((mat_p.(j).(i) /. row.(i)) *$ row)
    else if j = i then row
    else mat_p.(j)
  in

  let rec bw_subst_mat mat_pre n =
    let row = bw_subst_base mat_pre n (n+1) in
    let mat_cur = Array.init l (bw_subst_row mat_pre row n) in
    if n = 0 then mat_cur else bw_subst_mat mat_cur (n-1)
  in
      
  bw_subst_mat mat (l-1)


(* 連立方程式 : ax = b の解を計算する関数 *)
let solve a b =
  let l = Array.length a in
  let ext_coeff_elem p_a p_b i j = if j < l then p_a.(i).(j) else -. p_b.(i).(0) in
  let ext_coeff_row p_a p_b i = Array.init (l+1) (ext_coeff_elem p_a p_b i) in
  let ext_coeff_mat p_a p_b = Array.init l (ext_coeff_row p_a p_b) in

  let ab = ext_coeff_mat a b in
  let ab_t = fw_elim ab in
  let s_mat = bw_subst ab_t in
  Array.init n_deg (fun n -> [| -. (s_mat.(n).(l) /. s_mat.(n).(n)) |])  

(* 内積計算用関数 : ar2=横ベクトル, br2=縦ベクトル *)
let scalar n ar2 br2 =
  let c_arr = Array.init n_deg (fun m -> ar2.(n).(m) *. br2.(m).(0)) in
  Array.fold_left (+.) 0. c_arr

(* n列目のベクトルに対する 2乗ノルム計算関数 *)
let norm2 n ar2 =
  let c_arr = Array.init n_deg (fun m -> ar2.(m).(n) *. ar2.(m).(n)) in
  Array.fold_left (+.) 0. c_arr

(* newmark-β法の計算 : 現在値から次の加速度→速度→変位を計算 *)
let newmark_beta (m_ar2, c_ar2, k_ar2, a_m) fn (x_p, v_p, a_p) =
  (* 加速度を計算する内部関数 *)
  let acc x_pre v_pre fn =
    let v_tmp = Array.init n_deg (fun m -> [| v_p.(m).(0) +. dt *. (1. -. gamma) *. a_p.(m).(0) |]) in
    let x_tmp = Array.init n_deg (fun m -> [| x_p.(m).(0) +. dt *. v_p.(m).(0) +. 0.5 *. dt *. dt *. (1. -. 2.*.beta) *. a_p.(m).(0) |]) in
    let b_tmp = Array.init n_deg (fun m -> [| fn.(m).(0) -. (scalar m c_ar2 v_tmp) -. (scalar m k_ar2 x_tmp) |]) in
    solve a_m b_tmp
  in
  
  let a_c = acc x_p v_p fn in
  (* 加速度から速度と変位を計算 *)
  let v_c = Array.init n_deg (fun m -> [| v_p.(m).(0) +. dt *. ((1. -. gamma) *. a_p.(m).(0) +. gamma *. a_c.(m).(0))|]) in
  let x_c = Array.init n_deg (fun m -> [| x_p.(m).(0) +. dt *. v_p.(m).(0) +. 0.5 *. dt *. dt *. (1. -. 2.*.beta) *. a_p.(m).(0) +. beta *. dt *. dt *. a_c.(m).(0) |]) in
  (x_c, v_c, a_c)

(* 
   Newmark Beta 法の計算結果を表示 
   ~m : 質量行列
   ~c : 摩擦行列
   ~k : 剛性行列
   ~x0 : 位置初期条件
   ~v0 : 速度初期条件
   ~f_ext : 外力
   設定する順番は自由
*)
let newmark_beta_result ~m ~c ~k ~x0 ~v0 ~f_ext =
  let a_m = m +! ((dt *. gamma)*! c) +! ((dt *. dt *. beta) *! k) in

  (* 前回値→今回値をサンプル数分プロットする内部関数 *)
  let rec dump_newmark_beta n_cur f (x_p, v_p, a_p) =
    let t = (float_of_int n_cur) /. (float_of_int n_smp_1sec) in
    let (x_c, v_c, a_c) = newmark_beta (m, c, k, a_m) (f t) (x_p, v_p, a_p) in

    (* 結果のプロット *)
    print_string ((string_of_float t) ^ "\t");
    Array.iter (fun arr -> print_string ((string_of_float arr.(0)) ^ "\t")) x_c;
    print_endline "";
    if n_cur + 1 >= n_smp then () else dump_newmark_beta (n_cur+1) f (x_c, v_c, a_c)
  in

  (* 初期条件の設定 *)
  let f0 = f_ext 0. in
  let b0 =
    Array.init n_deg (fun i -> [| f0.(i).(0) -. (scalar i c v0) -. (scalar i k x0) |]) in
  (* 加速度の計算 *)
  let a0 = solve m b0 in

  (* 初期条件を設定し, 内部関数による計算 *)
  dump_newmark_beta 1 f_ext (x0,v0,a0)


(* main *)
let () =
  (* 質量行列 *)
  let m_ar2 =
    [|
      [| 1.; 0.; 0. |];
      [| 0.; 1.; 0. |];
      [| 0.; 0.; 1. |]
    |] in

  (* 剛性行列 *)
  let k_ar2 =
    [|
      [| 2000.; -1000.; 0. |];
      [|-1000.; 2000.; -1000. |];
      [| 0.; -1000.; 1000. |]
    |] in

  (* 摩擦抵抗 *)
  let c_ar2 =
    [|
      [| 5.; -5.; 0. |];
      [| -5.; 5.; 0. |];
      [| 0.; 0.; 0. |]
    |] in

  (* 微分方程式の初期条件 *)
  let x0 =
    [|
      [|0.|];
      [|0.|];
      [|0.|]
    |] in

  let v0 =
    [|
      [|0.|];
      [|0.|];
      [|0.|]
    |] in

  let ext_force t = Array.init n_deg (fun m -> if m = 2 && t <= 0.6 then [|10.|] else [|0.|]) in

  (*
    Newmark β 法による計算
    質量
    摩擦
    剛性
    初期条件(x0, v0)
    外力
    を登録して計算します
   *)
  newmark_beta_result ~m:m_ar2 ~c:c_ar2 ~k:k_ar2 ~x0:x0 ~v0:v0 ~f_ext:ext_force
