
open Lacaml.D

(* パイプライン演算子 *)
let (|>) x f = f x

(*  *)
let row_add u v =
  Array.init (Array.length u) (fun n -> u.(n) +. v.(n))
    
let row_sub u v =
  Array.init (Array.length u) (fun n -> u.(n) -. v.(n))

let row_time c v =
  Array.init (Array.length v) (fun n -> c *. v.(n))

let print_vector v =
  Array.iter (fun w -> print_string (string_of_float w)) v
    

(*  *)
let ( +! ) m1 m2 =
  Array.init (Array.length m1) (fun n -> row_add m1.(n) m2.(n))
let ( -! ) m1 m2 =
  Array.init (Array.length m1) (fun n -> row_sub m1.(n) m2.(n))
let ( *< ) c m =
  Array.init (Array.length m) (fun n -> row_time c m.(n))

let print_matrix arr2 =
  Array.iter (fun r -> print_vector r) arr2;
  print_endline ""


(* ax = b *)
let solve a b =
  let a_lm = lacpy (Mat.of_array a) in
  let b_lm = lacpy (Mat.of_array b) in
  gesv a_lm b_lm;
  Mat.to_array b_lm

(* bilinear form *)
let scalar n ar2 br2 =
  Array.init (Array.length br2) (fun m -> ar2.(n).(m) *. br2.(m).(0))
  |> Array.fold_left (+.) 0.

(* square norm *)
let norm2 n ar2 =
  Array.init (Array.length ar2) (fun m -> ar2.(m).(n) *. ar2.(m).(n))
  |> Array.fold_left (+.) 0.



let beta = 0.25
let gamma = 0.5

let n_smp_1sec = 100
let n_sec = 3
let n_smp = n_smp_1sec * n_sec
let dt = 1. /. (float_of_int n_smp_1sec)



(*
   Newmark beta method
 *)
let newmark_beta n_deg (m_ar2, c_ar2, k_ar2, a_m) fn (x_p, v_p, a_p) =
  let v_tmp = Array.init n_deg (fun m -> [| v_p.(m).(0) +. dt *. (1. -. gamma) *. a_p.(m).(0) |]) in
  let x_tmp = Array.init n_deg (fun m -> [| x_p.(m).(0) +. dt *. v_p.(m).(0) +. 0.5 *. dt *. dt *. (1. -. 2.*.beta) *. a_p.(m).(0) |]) in
  let b_tmp = Array.init n_deg (fun m -> [| fn.(m).(0) -. (scalar m c_ar2 v_tmp) -. (scalar m k_ar2 x_tmp) |]) in
  let a_c = solve a_m b_tmp in
  let v_c = Array.init n_deg (fun m -> [| v_p.(m).(0) +. dt *. ((1. -. gamma) *. a_p.(m).(0) +. gamma *. a_c.(m).(0))|]) in
  let x_c = Array.init n_deg (fun m -> [| x_p.(m).(0) +. dt *. v_p.(m).(0) +. 0.5 *. dt *. dt *. (1. -. 2.*.beta) *. a_p.(m).(0) +. beta *. dt *. dt *. a_c.(m).(0) |]) in
  (x_c, v_c, a_c)

(* 
   Dump results of Newmark beta
*)
let newmark_beta_result ~m ~c ~k ~x0 ~v0 ~f_ext =
	let n_deg = Array.length x0 in
  let a_m = m +! ((dt *. gamma)*< c) +! ((dt *. dt *. beta) *< k) in

  (*  *)
  let rec dump_newmark_beta n_cur f (x_p, v_p, a_p) =
    let t = (float_of_int n_cur) /. (float_of_int n_smp_1sec) in
    let (x_c, v_c, a_c) = newmark_beta n_deg (m, c, k, a_m) (f t) (x_p, v_p, a_p) in

    (* plot *)
    print_string ((string_of_float t) ^ "\t");
    Array.iter (fun arr -> print_string ((string_of_float arr.(0)) ^ "\t")) x_c;
    print_endline "";
    if n_cur + 1 >= n_smp then () else dump_newmark_beta (n_cur+1) f (x_c, v_c, a_c)
  in

  (* initial value *)
  let f0 = f_ext 0. in
  let b0 =
    Array.init n_deg (fun i -> [| f0.(i).(0) -. (scalar i c v0) -. (scalar i k x0) |]) in
  (*  *)
  let a0 = solve m b0 in

  (*  *)
  dump_newmark_beta 1 f_ext (x0,v0,a0)


(* main *)
let () =
  (*  *)
  let m_ar2 =
    [|
      [| 1.; 0.; 0. |];
      [| 0.; 1.; 0. |];
      [| 0.; 0.; 1. |]
    |] in

  (*  *)
  let k_ar2 =
    [|
      [| 2000.; -1000.; 0. |];
      [|-1000.; 2000.; -1000. |];
      [| 0.; -1000.; 1000. |]
    |] in

  (*  *)
  let c_ar2 =
    [|
      [| 5.; -5.; 0. |];
      [| -5.; 5.; 0. |];
      [| 0.; 0.; 0. |]
    |] in

  (*  *)
  let x0 =
    [|
      [|0.|];
      [|0.|];
      [|0.|]
    |] in

  let v0 =
    [|
      [|0.|];
      [|0.|];
      [|0.|]
    |] in

  let ext_force t = Array.init 3 (fun m -> if m = 2 && t <= 0.6 then [|10.|] else [|0.|]) in

  (*  *)
  newmark_beta_result ~m:m_ar2 ~c:c_ar2 ~k:k_ar2 ~x0:x0 ~v0:v0 ~f_ext:ext_force

