
let ( +@ ) z1 z2 = Complex.add z1 z2
let ( -@ ) z1 z2 = Complex.sub z1 z2
let ( *@ ) z1 z2 = Complex.mul z1 z2
let ( /@ ) z1 z2 = Complex.div z1 z2
		
let ( +< ) x z = { Complex.re = x +. z.Complex.re; Complex.im = z.Complex.im }
let i y = { Complex.re = 0.; Complex.im = y }
let ( *< ) x z = (x +< i 0.) *@ z
  

(*
 * 1 dimensional delay differential equation
 * from 
 * 
 *) 
module Basic=
  struct
    (* x''(t) + a1 x'(t) + a0 x(t) = bx(t - tau) *)
    type t = { a1 : float; a0 : float ; b : float; tau : float }

    (*
     * a = [| [| 0.; 1. |]; [| -.a0; -.a1 |] |] 
     * y' = map a y + mat b u
     *
     *)
    let pmatrix bs n =
      (* τ = (r + 1/2)h <-> h = τ / (r + 1/2) *)
      let h = bs.tau /. ((float n) +. 0.5) in
      let alpha = {
	Complex.re = -0.5 *. bs.a1;
	Complex.im = 0.5 *. sqrt (4. *. bs.a0 -. bs.a1 *. bs.a1)
      } in
      let s1 = alpha in
      let s2 = Complex.conj (alpha) in
      let e1 = Complex.exp (h *< s1) in
      let e2 = Complex.exp (h *< s2) in
      
      let p00 = s2 /@ (s2 -@ s1) *@ e1 -@ s1 /@ (s2 -@ s1) *@ e2 in
      let p01 = e2 /@ (s2 -@ s1) -@ e1 /@ (s2 -@ s1) in
      let p10 = (s1 *@ s2) /@ (s2 -@ s1) *@ e1 -@ (s1 *@ s2) /@ (s2 -@ s1) *@ e2 in
      let p11 = s2 /@ (s2 -@ s1) *@ e2 -@ s1 /@ (s2 -@ s1) *@ e1 in
      
      let r1 = (bs.b /. bs.a0) *< (1. +< (Complex.neg p00)) in
      let r2 = (bs.b /. bs.a0) *< p10 in
      (p00, p01, p10, p11, r1, r2)
	
    (* G-matrix *)
    let g_mat_c bs n =
      let (p00, p01, p10, p11, r1, r2) = pmatrix bs n in
      (* G-matrix の各成分 *)
      let g_mat_elem r i j =
	match i+1, j+1 with
	| 1,1 -> p00
	| 1,2 -> p01
	| 1,r -> r1
	| 2,1 -> p10
	| 2,2 -> p11
	| 2,r -> r2
	| 3,1 -> Complex.one
	| _ -> if i >= 3 && j = i - 1 then Complex.one else Complex.zero
      in
      Array.init (n+2) (fun k -> Array.init (n+2) (g_mat_elem (n+2) k))
  end

(*
 * Generic
 * dx(t)/dt = Ax(t) + Bu(t-τ)
 * u(t) = Dx(t)
 *
 *
 * dy(t)/dt = Ay(t) + Bv(t[i-r])
 * v(t[i]) = Dy(t[i])
 *)
