
(* SDOF における Semi-discretization *)

(*
  2 階微分方程式
  x''(t) + a1 x(t) + a0 x(t) = b0x(t-τ)
  を正規化する:
  dX(t)/dt = AX(t) + BX(t-τ)
 *)

(*
  パイプライン演算子
  OCaml 4.01 以降は定義不要
 *)
let (|>) x f = f x

(* 円周率 *)
let pi = 4. *. atan 1.

(* SDOFに関する 2 次元ベクトル, 2行列の演算 *)
let ( +& ) (x1, x2) (y1, y2) = (x1 +. y1, x2 +. y2)
let ( *< ) c (x1, x2) = (c *. x1, c *. x2)
let ( *& ) (a11, a12, a21, a22) (x1, x2) = 
  (a11 *. x1 +. a12 *. x2, a21 *. x1 +. a22 *. x2)
  
let ( +! ) (a11, a12, a21, a22) (b11, b12, b21, b22) =
  (a11 +. b11, a12 +. b12, a21 +. b21, a22 +. b22)
let ( -! ) (a11, a12, a21, a22) (b11, b12, b21, b22) =
  (a11 -. b11, a12 -. b12, a21 -. b21, a22 -. b22)
let ( *<! ) c (a11, a12, a21, a22) = (c*.a11, c*.a12, c*.a21, c*.a22)
let ( *! ) (a11, a12, a21, a22) (b11, b12, b21, b22) =
  (a11 *. b11 +. a12 *. b21, a11 *. b12 +. a12 *. b22, a21 *. b11 +. a22 *. b12, a12 *. b12 +. a22 *. b22)
let inv (a11, a12, a21, a22) =
  (1. /. (a11 *. a22 -. a12 *. a21)) *<! (a22, -.a21, -.a12, a11)
let ( /! ) (a11, a12, a21, a22) (b11, b12, b21, b22) =
  (a11, a12, a21, a22) *! (inv (b11, b12, b21, b22))
    
let e2 = (1., 0., 0., 1.)

(* exp の計算 *)
let exp_mat a1 a0 h =
  let p = 0.5 *. a1 in
  let q = a0 in
  let omega = sqrt (q -. p *. p) in
  let m11 = p *. exp (-.p *. h) /. omega *. sin (omega *. h) +. exp (-.p *. h) *. cos (omega *. h) in
  let m12 = exp (-.p *. h) /. omega *. sin (omega *. h) in
  let m21 = (-. q *. exp (-.p *. h) /. omega) *. sin (omega *. h) in
  let m22 = (-. p *. exp (-.p *. h) /. omega) *. sin (omega *. h) +. exp (-.p *. h) *. cos (omega *. h) in
  (m11,m12,m21,m22)

  
(*
  discrete_map ~tau:(2.*.pi) ~r:8 ~zeta:0. ~a0:1.5 ~b0:0.5
 *)
let discrete_map ~tau ~r ~zeta ~a0 ~b0 =
  let h = tau /. ((float r) +. 0.5) in
  let s = sqrt a0 in
  let a1 = 2. *. zeta *. s in
  let am = (0., 1., -.a0, -.a1) in
  let bm = (0., b0) in
  let (p11, p12, p21, p22) = exp_mat a1 a0 h in
  let (r1, r2) =  ((exp_mat a1 a0 h -! e2) /! am) *& bm in 
  let elem i j =
    if (i+1,j+1) = (1,1) then p11
    else if (i+1,j+1) = (1,2) then p12
    else if (i+1,j+1) = (2,1) then p21
    else if (i+1,j+1) = (2,2) then p22
    else if (i+1,j+1) = (1,r+2) then r1
    else if (i+1,j+1) = (2,r+2) then r2
    else if (i+1,j+1) = (3,1) then 1.
    else if (i+1 >= 4) && i-1 = j then 1.
    else 0.
  in
  Array.init (r+2) (fun l -> Array.init (r+2) (elem l))

(* 線形写像  *)
let lmap g v =
  let dim = Array.length g in
  let w n = Array.init dim (fun k -> g.(n).(k) *. v.(k).(0)) |> Array.fold_left (+.) 0. in
  Array.init dim (fun l -> [| w l |])

(* 初期条件を設定 *)
let init_config x0 v0 x_pre =
  let dim = Array.length x_pre in
  let elem k =
    match k+1 with
    | 1 -> [| x0 |]
    | 2 -> [| v0 |]
    | _ -> [| x_pre.(k-2).(0) |]
  in
  Array.init (dim+2) elem

let nsmp = 32
let g = discrete_map ~tau:(2.*.pi) ~r:nsmp ~zeta:0.05 ~a0:1.0 ~b0:0.1

open Lazy
exception Empty_stream
type 'a stream = Nils | Cons of 'a * 'a stream Lazy.t

(* semidiscrete の遅延評価 *)
let rec lazy_semidiscrete q0 =
  Cons (q0, lazy (lazy_semidiscrete (lmap g q0)))

let rec stream_ref s n =
  match s with
  | Nils -> raise Empty_stream
  | Cons (q, _) when n = 1 -> q
  | Cons (_, tail) -> stream_ref (force tail) (n-1)
    
let q0 = init_config 0. 1. (Array.make_matrix nsmp 1 0.)

let s = lazy_semidiscrete q0
let print_position q = print_float q.(0).(0)
let rec dump_dde_stream n idx =
  print_position (stream_ref s (idx+1));
  print_endline "";
  if (idx + 1) > n then ()
  else dump_dde_stream n (idx+1)

let () = dump_dde_stream 1000 0
