open Format
open Bigarray

open Lacaml.C
open Lacaml.Io


let pi = 4. *. atan 1.

(* 実数のべき乗 *)
let rec pow a_pre a_base n =
  if n = 0 then a_pre
  else pow (a_pre *. a_base) a_base (n-1)

let ( *^ ) a n = pow 1. a n

let i y = { Complex.re = 0.; Complex.im = y }
let ( +< ) x z =
  let xc = { Complex.re = x; Complex.im = 0. } in
  Complex.add xc z

    

(* 横ベクトルの計算 *)
let plus_row u v =
  let dim = Array.length u in
  Array.init dim (fun n -> Complex.add u.(n) v.(n))

let minus_row u v =
  let dim = Array.length u in
  Array.init dim (fun n -> Complex.sub u.(n) v.(n))

let prod_row c v =
  let dim = Array.length v in
  Array.init dim (fun n -> Complex.mul c v.(n))

let re { Complex.re; Complex.im } = re
let im { Complex.re; Complex.im } = im


(* 行列の計算 *)
let ( +! ) m1 m2 =
  let dim = Array.length m1 in
  Array.init dim (fun n -> plus_row m1.(n) m2.(n))

let ( -! ) m1 m2 =
  let dim = Array.length m2 in
  Array.init dim (fun n -> minus_row m1.(n) m2.(n))

let ( *< ) c m =
  let dim = Array.length m in
  Array.init dim (fun n -> prod_row c m.(n))

let scalar u v i j =
  let l = Array.length u.(0) in
  let uv = Array.init l (fun n -> Complex.mul u.(i).(n) v.(n).(j)) in
  Array.fold_left Complex.add Complex.zero uv

(* 行列の積 *)
let ( *! ) a b =
  let row = Array.length a in
  let col = Array.length b.(0) in
  let ab_row k = Array.init col (fun n -> scalar a b k n) in
  Array.init row ab_row


(* 刃の枚数 *)
let z = 2

let m_x = 10.
let m_y = 10.

let k_x = ((2. *. pi) *^ 2) *. 10. ** 7.
let k_y = ((2. *. pi) *^ 2) *. 10. ** 7.

let c_x = 300.
let c_y = 300.

(* 固有振動数  *)
let wn_x = sqrt (k_x /. m_x)
let wn_y = sqrt (k_y /. m_y)

let kt = 1500. *. 10. ** 6.
let kr = 0.3

(* A-matrix *)
let a0 ths the =
  let n_4pi = (float_of_int z) /. (4. *. pi) in
  let (cse, sne) = (cos (2. *. the), sin (2. *. the)) in
  let (css, sns) = (cos (2. *. ths), sin (2. *. ths)) in
  [| [| (n_4pi *. ((-.sne-.2.*.the+.kr*.cse) -. (-.sns-.2.*.ths+.kr*.css))) +< i 0.;
       (n_4pi *. ((-.sne-.2.*.the+.kr*.cse) -. (-.sns-.2.*.ths+.kr*.css))) +< i 0. |];
    [| (n_4pi *. ((-.sne+.2.*.the+.kr*.cse) -. (-.sns+.2.*.ths+.kr*.css))) +< i 0. ;
      (n_4pi *. ((-.sne+.2.*.the+.kr*.cse) -. (-.sns+.2.*.ths+.kr*.css))) +< i 0. |] |]

let we = if wn_x > wn_y then 2. *. wn_x else 2. *. wn_y

let eigen_values m_arr =
  let m = Mat.of_array m_arr in
  let m_copy = lacpy m in
  geev m_copy

let plot_stability_point k w eig =
  let l_re = re (Complex.neg (Complex.inv eig)) in
  let l_im = im (Complex.neg (Complex.inv eig)) in
  let alim = -. 1000. *. l_re /. kt *. (1. +. ((l_im /. l_re) *^ 2)) in
  let ep = pi -. 2. *. atan (l_im /. l_re) in
  if 0. < alim && alim < 50. then
    begin
      let n_rpm = 60.0 *. w /. ((float_of_int z) *. (2. *. (float_of_int (k-1)) *. pi +. ep)) in
      print_endline ((string_of_float n_rpm) ^ " " ^ (string_of_float alim))
    end
      
(* plot stability diagram *)
let rec stability_diagram k f0 n =
  let w = 2. *. pi *. ((float_of_int f0) +. (float_of_int n) *. 0.05) in
  let g = [| [| Complex.inv ((k_x -. m_x *. (w *^2)) +< i (c_x *. w)); Complex.zero |];
	    [| Complex.zero ; Complex.inv ((k_y -. m_y *. (w *^2)) +< i (c_y *. w)) |] |] in
  let (_, eig_v, _) = eigen_values (a0 0. (0.25 *. pi) *! g) in
  Array.iter (plot_stability_point k w) (Vec.to_array eig_v);
  (* tail recursion *)
  if 2. *. pi *. ((float_of_int f0) +. (float_of_int (n+1)) *. 0.05) >= we then ()
  else stability_diagram k f0 (n+1)

(* entry point *)
let () =
  for deg = 1 to 5 do
      stability_diagram deg 980 0;
  done


