(* 
 Excel による数値解析
 [北陸電力ミックスオイラープロジェクトチーム著]
 P. 129 の連立微分方程式 (5.20), (5.21) を計算

 ミックスオイラー法に関する利用上の注意 : 
 ミックスオイラー法を用いた数値解析装置及び数値解析プログラムは,
 北陸電力株式会社が日本および米国内で特許出願中であり, 
 商用目的等での使用を希望される場合は, 事前に北陸電力株式会社までご連絡下さい. 
   http://www.rikuden.co.jp/products/shouhin19.html
 *)

(* mixed euler 法で次のステップの値を計算 *)
let mixed_euler_next h a ((phi_p, delta_p), (dot_phi_p, dot_delta_p)) =
  (* 一般化速度 : 前進 Euler 法で次のステップを計算 *)
  let (ddot_phi, ddot_delta) = a ((phi_p, delta_p), (dot_phi_p, dot_delta_p)) in
  let dot_phi_n = dot_phi_p +. h *. ddot_phi in
  let dot_delta_n = dot_delta_p +. h *. ddot_delta in
  (* 一般化座標 : 一般化速度を使って後退 Euler 法の計算を適用 *)
  let phi_n = phi_p +. h *. dot_phi_n in
  let delta_n = delta_p +. h *. dot_delta_n in
  ((phi_n, delta_n), (dot_phi_n, dot_delta_n))

open Lazy
exception Empty_stream

(* 遅延ストリーム *)
type 'a stream = Nils | Cons of 'a * 'a stream Lazy.t

(* mixed-euler 法による遅延ストリーム定義 *)
let rec lazy_mixed_euler h a ((phi_0, delta_0), (dot_phi_0, dot_delta_0)) = Cons (((phi_0, delta_0), (dot_phi_0, dot_delta_0)), lazy (lazy_mixed_euler h a (mixed_euler_next h a ((phi_0, delta_0), (dot_phi_0, dot_delta_0)))))

let tpi = 8. *. atan 1.
let omega = tpi *. 60.

(* (5.20), (5.21) 式の計算 : 加速度算出 *)
let a_turbine ((phi, delta), (dot_phi, dot_delta)) = 
  let ig = 10. in
  let it = 9. in
  let k = 44000. in
  let ve = 500. in
  let vt = 200. in
  let zi = 0.5 in
  let pt = 500. in

  let ddot_phi = (pt /. (dot_phi +. omega) -. k *. (phi -. delta)) /. it in
  let ddot_delta = (-. (ve *. vt *. (sin delta)) /. (zi *. (dot_delta +. omega)) +. k *. (phi -. delta)) /. ig in

  (ddot_phi, ddot_delta)


(* 遅延ストリームから n 番目の値を取り出す *)
let rec stream_ref s n =
  match s with
  | Nils -> raise Empty_stream
  | Cons (x, _) when n = 1 -> x
  | Cons (_, tail) -> stream_ref (force tail) (n-1)

(*
 微分方程式の計算をする遅延ストリーム
 初期条件 : φ0 = 0.5, δ0 = 0.6; 速度は共に 0 
*)
let s = lazy_mixed_euler 0.001 a_turbine ((0.5, 0.6), (0., 0.))
(* δの値を出力 *)
let print_delta ((phi, delta), (dot_phi, dot_delta)) = print_float delta
(* idx 番目から n 番目までの値を主力する *)
let rec dump_stream n idx =
  print_delta (stream_ref s (idx+1));
  print_endline "";
  if (idx + 1) > n then ()
  else dump_stream n (idx+1)

(* エントリポイント *)
let () = dump_stream 3000 0
